local _, class = UnitClass("player")
if class ~= "DEATHKNIGHT" then
	return
end

mRunes = LibStub("AceAddon-3.0"):NewAddon("mRunes", "AceEvent-3.0", "AceConsole-3.0")
local mRunes = mRunes

local LSM 	= LibStub("LibSharedMedia-3.0")
local FA	= LibStub("LibFrameAnchorRegistry-1.0")
LSM:Register("font", "accid", [[Interface\AddOns\mRunes\accid.ttf]])
LSM:Register("statusbar", "HalD", [[Interface\AddOns\mRunes\HalD.tga]])

-- Locals
local db
local runicPowerBar
local FontFlagList = {OUTLINE = "Outline", THICKOUTLINE = "Thick Outline", NONE = "None", MONOCHROME = "Monochrome"}

-- Rune Types
local RUNETYPE_BLOOD = 1;
local RUNETYPE_UNHOLY = 2;
local RUNETYPE_FROST = 3;
local RUNETYPE_DEATH = 4;

-- GUI
local UIF
local AceGUI = LibStub("AceGUI-3.0")

function mRunes:OnInitialize()
	local defaults = {
		profile = {			
			Enabled = true,		
			Width = 100,
			Height = 100,
			Show = false,
			x = 0, 
			y = -100,
		}
	}
	
	self.db  = LibStub("AceDB-3.0"):New("mRunesDB", defaults, "Default")
	self.db.RegisterCallback(self, "OnProfileChanged", "ProfileChanged")
	self.db.RegisterCallback(self, "OnProfileCopied", "ProfileChanged")
	self.db.RegisterCallback(self, "OnProfileReset", "ProfileChanged")
	db = self.db.profile
	
	self.Callbacks = LibStub("CallbackHandler-1.0"):New(self)
	
	self:RegisterChatCommand("mr", "ToggleConfig")	
	self:SetupLDB()	
	self:SetEnabledState(db.Enabled)
	
	local UIFSettings = {
		font = LibStub("LibSharedMedia-3.0"):Fetch("font", "accid"),
		fontSize = 16,
		Text2 = {
			fontSize = 14,
			fontColor = {1, 0.2, 0.2},
		}
	}

	UIF = LibStub("LibGUIFactory-1.0"):GetFactory("mRunes", UIFSettings)
end

function mRunes:OnEnable()
	if not self.MainFrame then
		self:CreateMainFrame()
	end

	-- Register Events	
	self:RegisterEvent("PLAYER_REGEN_DISABLED", "EnterCombat")
	self:RegisterEvent("PLAYER_REGEN_ENABLED", "LeaveCombat")		
	
	FA.RegisterCallback(self, "FrameRegistered", "ApplySettings")
	
	self:ApplySettings()
end

function mRunes:OnDisable()
end

function mRunes:ProfileChanged()
	db = self.db.profile
	self:ApplySettings()
end

-- ~~~~~~~~~~~~~~~~~~~~~~
-- Events ~~~~~~~~~~~~~~~
-- ~~~~~~~~~~~~~~~~~~~~~~
function mRunes:EnterCombat(event)
	self.Callbacks:Fire("EnteringCombat")
end

function mRunes:LeaveCombat(event)
	self.Callbacks:Fire("LeavingCombat")
end

-- ~~~~~~~~~~~~~~~~~~~~~~
-- Frame ~~~~~~~~~~~~~~~~
-- ~~~~~~~~~~~~~~~~~~~~~~
function mRunes:CreateMainFrame()
	local frame = CreateFrame("Frame", nil, UIParent)
	local t = frame:CreateTexture()
	t:SetAllPoints()
	t:SetTexture(0, 1, 1, 0.5)
	frame.bg = t
	
	FA:RegisterAnchor("mRunes", "Center", "mRunesCenter", frame)
	
	local text = frame:CreateFontString(nil)
	text:SetFont(LSM:Fetch("font", "accid"), 20, "OUTLINE")
	text:SetText("mRunes Center")
	text:SetAllPoints()
	
	frame:Hide()
	
	self.MainFrame = frame
end

-- ~~~~~~~~~~~~~~~~~~~~~~
-- Settings ~~~~~~~~~~~~~
-- ~~~~~~~~~~~~~~~~~~~~~~
function mRunes:ApplySettings()
	self.MainFrame:ClearAllPoints()
	self.MainFrame:SetPoint("CENTER", db.x, db.y)
	self.MainFrame:SetWidth(db.Width)
	self.MainFrame:SetHeight(db.Height)
	
	if db.Show then
		self.FadeIn(self.MainFrame)
	else
		self.FadeOut(self.MainFrame)
	end
end


-- ~~~~~~~~~~~~~~~~~~~~~~~
-- Anchoring ~~~~~~~~~~~~~
-- ~~~~~~~~~~~~~~~~~~~~~~~
do
	local AnchorList = {}
	
	function mRunes:GetAnchorList()		
		AnchorList = AnchorList or {}
		AnchorList["TOPLEFT"] = "Top left"
		AnchorList["TOP"] = "Top"
		AnchorList["TOPRIGHT"] = "Top Right"
		AnchorList["RIGHT"] = "Right"
		AnchorList["BOTTOMRIGHT"] = "Bottom right"
		AnchorList["BOTTOM"] = "Bottom"
		AnchorList["BOTTOMLEFT"] = "Bottom left"
		AnchorList["LEFT"] = "Left"
		AnchorList["CENTER"] = "Center"
		
		return AnchorList
	end
end


-- ~~~~~~~~~~~~~~~~~~~~~~
-- Config ~~~~~~~~~~~~~~~
-- ~~~~~~~~~~~~~~~~~~~~~~
do	
	local configMode = false
	local function ToggleConfigMode()
		configMode = not configMode
		mRunes.Callbacks:Fire("ToggleConfigMode", configMode)
	end

	local function Callback()
		mRunes:ApplySettings()
	end
	
	local SortedModules = {}
	local OptionsStatus = {}
	local OptionsFrame
	
	local function SortModules(a, b)
		return a.DisplayOrder < b.DisplayOrder
	end

	function mRunes:ToggleConfig()
		if not OptionsFrame then	
			OptionsFrame = UIF:Frame("mRunes")
			OptionsFrame:SetLayout("Flow")
			OptionsFrame:SetWidth(813)
			OptionsFrame:SetHeight(630)
			OptionsFrame:Hide()			
			OptionsFrame:SetCallback("OnClose", function(widget)				
				OptionsFrame:ReleaseChildren()		
				OptionsFrame.frame:Hide()
			end)			
		end
		
		if OptionsFrame.frame:IsShown() then
			OptionsFrame:Hide()
			return
		else			
			OptionsFrame:AddChild(UIF:Button("Toggle Config Mode", ToggleConfigMode))
			m = UIF:TreeGroup()
			m:SetFullHeight(true)
			m:SetLayout("Fill")	
			m:SetCallback("OnGroupSelected", function(self, event, group) 
				SelectedOption = group 
				self:DoLayout()
				self.parent:DoLayout()
			end)									
			OptionsFrame:AddChild(m)
		end
		
		m:AddGroup({{text = "Main", value = "Main"}}, function()
			local returnGroup = AceGUI:Create("SimpleGroup")
			returnGroup:SetLayout("Flow")
			returnGroup:SetFullWidth(true)
			
			returnGroup:AddChild(UIF:CheckBox("Show mRunes Center", db, "Show", Callback, "The center is the point hat all modules position around"))
			
			local group = UIF:InlineGroup1("Position")
			returnGroup:AddChild(group)			
			group:AddChild(UIF:Text1("This position will effect all modules. Use to move mRunes around as a whole when you have positioned each item relative to eachother as you want them."))
			group:AddChild(UIF:PositionSlider("x", "X", db, "x", 0.1, Callback, 0.5))
			group:AddChild(UIF:PositionSlider("y", "Y", db, "y", 0.1, Callback, 0.5))
			
			return returnGroup
		end)
		
		wipe(SortedModules)
		for moduleName, module in self:IterateModules() do
			if module.GetOptions and type(module.GetOptions) == "function" then
				tinsert(SortedModules, module)
			end
		end
		table.sort(SortedModules, SortModules)
		
		for i, module in pairs(SortedModules) do
			if module.GetOptions and type(module.GetOptions) == "function" then
				m:AddGroup({{text = module.DisplayName or module:GetName(), value = module:GetName()}}, module:GetOptions())
			end
		end
		
		m:AddGroup({{text = "Profile", value = "Profile"}}, function()
			return UIF:AceProfileOption(mRunes.db)
		end)
		
		m:SetSelected(SelectedOption or "Main")
		m:SetStatusTable(OptionsStatus)	
		OptionsFrame:Show()
	end

	function mRunes:SetupLDB()
		local LDB = LibStub("LibDataBroker-1.1")
		if not LDB then return end
		
		local l = LDB:NewDataObject("mRunes")
		l.type = "launcher"
		l.icon = [[Interface\Icons\Spell_DeathKnight_RuneTap]]
		l.tocname = "mRunes"
		l.label = "mRunes"
		l.OnClick = function(self, button)
			mRunes:ToggleConfig()
		end
		l.OnTooltipShow = function(tt)
			tt:AddLine("mRunes")
			tt:AddLine("Click to toggle configuration")
		end
		
		self.ldbojb = l
	end
end

-- OnUpdate Stuff -------------------
-------------------------------------
do
	local OnUpdateFuncs = {}
	local FuncElapsed = {}
	local UpdateFrame = CreateFrame("Frame")
	UpdateFrame:SetScript("OnUpdate", function(self, elapsed)
		for module, functions in pairs(OnUpdateFuncs) do
			if functions then
				for func, data in pairs(functions) do
					if data.updateRate then
						data.elapsed = data.elapsed + elapsed
						if data.elapsed > data.updateRate then
							func(module, data.elapsed)
							data.elapsed = 0
						end
					else
						func(module, elapsed)
					end
				end
			end
		end
	end)	
	
	function mRunes:RegisterOnUpdate(module, func, updateRate)
		OnUpdateFuncs[module] = OnUpdateFuncs[module] or {}
		OnUpdateFuncs[module][func] = {updateRate = updateRate, elapsed = 0}
	end
	
	function mRunes:UnregisterOnUpdate(module, func)
		if OnUpdateFuncs[module] then
			OnUpdateFuncs[module][func] = nil
		end
	end
	
	local function Fader(self, elapsed)
		local currentAlpha = self:GetAlpha()
		if self.fadingIn then
			self:Show()
			if currentAlpha >= self.fadeInTarget then 
				mRunes:UnregisterOnUpdate(self, Fader)
				self:SetAlpha(self.fadeInTarget)
				self.fadeInTarget = nil
				self.fadingIn = false				
				return
			end
			self:SetAlpha(self:GetAlpha() + elapsed * 2)
		else
			if currentAlpha <= self.fadeOutTarget then
				self:SetAlpha(self.fadeOutTarget)
				if self.fadeOutTarget <= 0 then
					self:Hide()					
				end
				self.fadeOutTarget = nil
				mRunes:UnregisterOnUpdate(self, Fader)
				return
			end
			self:SetAlpha(self:GetAlpha() - elapsed * 2)
		end
	end
	
	function mRunes.FadeIn(frame, reset, target)
		frame.fadingIn = true
		frame.fadeInTarget = target or 1
		if reset then frame:SetAlpha(0) end
		mRunes:RegisterOnUpdate(frame, Fader)
	end
	
	function mRunes.FadeOut(frame, target)
		frame.fadingIn = nil
		frame.fadeOutTarget = target or 0
		mRunes:RegisterOnUpdate(frame, Fader)
	end
end